/*
 * Copyright (C) Sergey P. Derevyago, 2005.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

#include "key.hpp"
#include <string.h>
#include "derscrypt.hpp"

using namespace std;

Key::Key(const void* ptr, int size)
{
 bytes.resize(size);
 memcpy(&bytes[0], ptr, size);

 checkKey(&bytes[0], bytes.size());

 const int num=9;  // last is value^256
 powers.reserve(num);

 powers.push_back(newUInt(ptr, size));
 for (int i=1; i<num; i++)
     powers.push_back(powers[i-1]*powers[i-1]);
}

sh_ptr<vector<ShUInt> > Key::computeRemainders(ShUInt a)
{
 int depth;
 for (depth=-1; a>=powers[depth+1]; )
     depth++;

 sh_ptr<vector<ShUInt> > ret(new vector<ShUInt>);
 if (depth>=0) ret->reserve(1<<depth+1);

 computeRemainders(a, depth, *ret);
 return ret;
}

void Key::computeRemainders(ShUInt a, int depth, vector<ShUInt>& rems)
{
 if (depth<0) {
    rems.push_back(a);
    return;
 }

 ShUInt q=newUInt(), r=newUInt();
 UInt::qr(*a, *powers[depth], *q, *r);

 if (depth==0) {
    if (*q!=UInt::zero || rems.size()>0) rems.push_back(q);
    if (*r!=UInt::zero || rems.size()>0) rems.push_back(r);

    return;
 }

 computeRemainders(q, depth-1, rems);
 computeRemainders(r, depth-1, rems);
}

const vector<int>& Key::getPermutation(int length)
{
 for (int i=0, end=cache.size(); i<end; i++) {
     if (int(cache[i]->size())==length)
        return *cache[i];
 }

 sh_ptr<vector<int> > perm=createPermutation(length, getValue());
 cache.push_back(perm);

 return *perm;
}

sh_ptr<vector<int> > Key::createPermutation(int len, ShUInt b)
{
 ShUInt b2=newUInt(*b);
 sh_ptr<vector<int> > ret(new vector<int>(len)), val(new vector<int>(len));

 int med=len-len/2;
 for (int i=0; i<len; i++) {
     (*ret)[i]=-2;

     int pos= (i<med) ? len-med+i : i-med;
     (*val)[pos]=i;
 }

 int points[]={0, len-med, len};
 for (int p=0; p<2; p++) {
     int pbeg=points[p], plen=points[p+1]-pbeg;

     for (int i=0, sub=0; i<plen; i++) {
         if (i==plen-2 && sub==0) {  // don't left last element alone
            for (int j=0; ; j++) {
                if ((*ret)[pbeg+j]==-2) {
                   if ((*ret)[pbeg+j+1]==-2) {
                      (*ret)[pbeg+j]=(*val)[pbeg+i+1];
                      (*ret)[pbeg+j+1]=(*val)[pbeg+i];
                      goto out;
                   }

                   break;
                }
            }
         }

         UInt div(plen-i-sub);
         if (*b2<div) *b2=*b;  // restart from b 

         UInt q, r;
         UInt::qr(*b2, div, q, r);
         *b2=q;
         int rem=r.toUInt();

         for (int j=0, k=-1, v=(*val)[pbeg+i]; ; j++) {
             if ((*ret)[pbeg+j]==-2 && !(j>0 && v==(*ret)[pbeg+j-1]+1) &&
                 ++k==rem) {
                (*ret)[pbeg+j]=v;
                sub= (j+1<plen && (*ret)[pbeg+j+1]==-2) ? 1 : 0;

                break;
             }
         }
     }
     out:;
 }

 return ret;
}

