/*
 * Copyright (C) Sergey P. Derevyago, 2005.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Defines key for DersCrypt cipher and provides key-dependant functionality.
 */
 
#ifndef __KEY_HPP__
#define __KEY_HPP__

#include <vector>
#include "uint.hpp"

class Key {
      std::vector<uchar> bytes;
      std::vector<ShUInt> powers;
      std::vector<sh_ptr<std::vector<int> > > cache;

      // forbidden copying
      Key(const Key&);
      Key& operator=(const Key&);

      void computeRemainders(ShUInt a, int depth, std::vector<ShUInt>& rems);
      sh_ptr<std::vector<int> > createPermutation(int len, ShUInt b);

 public:
      Key(const void* ptr, int size);

      void* operator new(size_t) { return fixed_alloc<Key>::alloc(); }
      void operator delete(void* ptr, size_t) { fixed_alloc<Key>::free(ptr); }

      const std::vector<uchar>& getBytes() const { return bytes; }
      const ShUInt& getValue() const { return powers[0]; }
      sh_ptr<std::vector<ShUInt> > computeRemainders(ShUInt a);
      const std::vector<int>& getPermutation(int length);
};

inline sh_ptr<Key> newKey(const void* ptr, int size)
{
 return sh_ptr<Key>(new Key(ptr, size));
}

#endif

